/*****************************************************************************
 *
 *   h264_common.cpp
 *
 *   H.264 v3 interface software fallback implementation - shared code,
 *
 *   $Id: //icaclient/unix13.3/client/linux/obj.h264/h264_common.cpp#1 $
 *
 *   Copyright 2013-2015 Citrix Systems, Inc.  All Rights Reserved.
 *
 *****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include <execinfo.h>
#include <unistd.h>
#include <time.h>
#include <dlfcn.h>

#include <algorithm>

#include <sys/shm.h>
#include <sys/ipc.h>

#ifdef X11_SUPPORT
#include <X11/X.h>
#endif /* X11_SUPPORT */

#include <linux/limits.h>

#include "citrix.h"
#include "UI_if.h"
#include "ctxh264.h"
#include "mainloop.h"

#ifndef PLUGIN_SDK
#include "hostfeat.h"
#endif

#ifdef X11_SUPPORT
/* Xv* function prototypes. */

typedef XvImage
    *(*pfn_XvShmCreateImage)(Display *dpy, XvPortID port, int id, char *data,
			     int width, int height, XShmSegmentInfo *shminfo);
typedef int
    (*pfn_XvShmPutImage)(Display *dpy, XvPortID port, Drawable d, GC gc,
			 XvImage *image, int src_x, int src_y,
			 unsigned int src_w, unsigned int src_h,
			 int dest_x, int dest_y,
			 unsigned int dest_w, unsigned int dest_h,
			 Bool send_event);
typedef int
    (*pfn_XvQueryAdaptors)(Display *dpy, Window window,
			   unsigned int *p_num_adaptors,
			   XvAdaptorInfo **pp_adaptor_info);
typedef void (*pfn_XvFreeAdaptorInfo)(XvAdaptorInfo *p_adaptor_info);
typedef XvImageFormatValues
    *(*pfn_XvListImageFormats)(Display *dpy, XvPortID port, int *count);
typedef int (*pfn_XvGrabPort)(Display *dpy, XvPortID port, Time time);

/* Pointers to functions after dynamically loading libXv. */

static pfn_XvShmCreateImage   _XvShmCreateImage = NULL;
static pfn_XvShmPutImage      _XvShmPutImage = NULL;
static pfn_XvQueryAdaptors    _XvQueryAdaptors = NULL;
static pfn_XvFreeAdaptorInfo  _XvFreeAdaptorInfo = NULL;
static pfn_XvListImageFormats _XvListImageFormats = NULL;
static pfn_XvGrabPort         _XvGrabPort = NULL;

static Display               *disp = NULL;
static Screen                *scr = NULL;
static bool                   g_bUsingXVideoExtension = false;
static bool                   g_bUsingXShmExtension = false;

#endif /* X11_SUPPORT */

static GstContextMap g_GstCtxMap;
static UINT32 g_GstDecodeCtxListIdx = H264_INVALID_CONTEXT + 1;

V3FBCtxData *GetCtxData(H264_context Ctx)
{
    V3FBCtxData *pCtxData;

    /* map.at () function throws if element is not present at the location
     * specified by the key.
     */

    pCtxData = g_GstCtxMap.at(Ctx);
    return pCtxData;
}

/* This function would be called only once, to initialize the DLL. */

#ifdef X11_SUPPORT
bool X11_init(void)
{
    /* Check for at least 24-bit colour depth. This is done here as
     * returning "false" will result in Receiver falling back to
     * compatible (JPEG) mode.
     */

    /* Set up X11 display variables. */

    disp = GetICADisplay(); /* Exported by the main Receiver process. */
    scr = DefaultScreenOfDisplay(disp);

    if (DefaultDepth(disp, DefaultScreen(disp)) < 24) {
        /* This decoder can't deal with anything other than 24 or 32-bit
         * displays.
         */
        /* FIX ME! */

        return false;
    }
    return true;
}
#endif /* X11_SUPPORT */

void close_context_common(V3FBCtxData *pCtxData)
{
    if (pCtxData) {
        pthread_mutex_lock(&pCtxData->H264DataLock);

        if (pCtxData->pEncodedFrame) {
            free(pCtxData->pEncodedFrame);
        }

#ifdef X11_SUPPORT
        if (g_bUsingXVideoExtension || g_bUsingXShmExtension) {
            XShmDetach(disp, &pCtxData->shm_info);
            if (g_bUsingXVideoExtension) {
                if (pCtxData->fb_yuv) {
                    XFree(pCtxData->fb_yuv);
                    pCtxData->fb_yuv = 0;
                }
            }
            /* Free shared memory. */

            shmdt(pCtxData->shm_info.shmaddr);
            shmctl(pCtxData->shm_info.shmid, IPC_RMID, 0);
        }

        if (pCtxData->fb) {
            /* XDestroyImage also frees pCtxData->fb->data */

            XDestroyImage(pCtxData->fb);
        }
#endif /* X11_SUPPORT */

        pthread_mutex_unlock(&pCtxData->H264DataLock);
        pthread_mutex_destroy(&pCtxData->H264DataLock);

#ifdef DUMP_OUTPUT
        if (pCtxData->fp) {
            fclose(pCtxData->fp);
        }
#endif
        delete pCtxData;
    }
}

void end_common(void)
{
    V3FBCtxData *pCtxData;

    /* Close all contexts. */

    for (GstContextMap::iterator i = g_GstCtxMap.begin();
	 i != g_GstCtxMap.end();
	 i++) {
        pCtxData = i->second;
        close_context(pCtxData);
    }
}

V3FBCtxData *open_context_common(int width, int height, void *codec_data,
				 int len, unsigned int options)
{
    static bool  init = false;
    H264_context CurCtx = g_GstDecodeCtxListIdx;
    V3FBCtxData *pCtxData;
    UINT32       size = 0;

    REFERENCE_VAR(codec_data);
    REFERENCE_VAR(len);
    REFERENCE_VAR(options);

    pCtxData = new V3FBCtxData;
    pCtxData->CtxId = CurCtx;
    pCtxData->uCtxWidth	= width;
    pCtxData->uCtxHeight = height;
    pCtxData->uEncMaxFrameSize = 0;
    pCtxData->pEncodedFrame = NULL;

    pCtxData->fb = 0;
    pCtxData->fb_yuv = 0;
    pCtxData->ai = 0;

    pthread_mutex_init(&pCtxData->H264DataLock, NULL);

#ifdef X11_SUPPORT
    if (!init) {
        /* Perform one time initialization post Thinwire init here, where we
         * know whether Receiver is configured to use SHM or not.
         */

        init = true;
        if (getenv("CTXH264_FB_USEXVIDEO")) {
            int major_op, first_event, first_error;

            /* Check to see if the XVideo extension is supported.
	     * If so, disable seamless and lossless text support as this
	     * is currently unsupported with the XVideo extension.
             */

            if (XQueryExtension(disp, "XVideo", &major_op,
				&first_event, &first_error)) {
                void *libXv = dlopen("libXv.so", RTLD_NOW);
                if (!libXv) {
                    libXv = dlopen("libXv.so.1", RTLD_NOW);
                }
                if (libXv) {
                    /* Get Xv function pointers. */
                    _XvShmCreateImage =
		       (pfn_XvShmCreateImage)dlsym(libXv, "XvShmCreateImage");
                    _XvShmPutImage =
		       (pfn_XvShmPutImage)dlsym(libXv, "XvShmPutImage");
                    _XvQueryAdaptors =
		       (pfn_XvQueryAdaptors)dlsym(libXv, "XvQueryAdaptors");
                    _XvFreeAdaptorInfo =
		       (pfn_XvFreeAdaptorInfo)dlsym(libXv,
						    "XvFreeAdaptorInfo");
                    _XvListImageFormats =
		       (pfn_XvListImageFormats)dlsym(libXv,
						     "XvListImageFormats");
                    _XvGrabPort = (pfn_XvGrabPort)dlsym(libXv, "XvGrabPort");
                }
            }
            g_bUsingXVideoExtension = _XvShmCreateImage && _XvShmPutImage 
                            && _XvQueryAdaptors && _XvFreeAdaptorInfo
                            && _XvListImageFormats && _XvGrabPort;

            if (g_bUsingXVideoExtension) {
                /* XVideo supported.
		 * Disable lossless text and seamless support.
		 */

	        H264_decoder.options &= ~(H264_OPTION_LOSSLESS |
					  H264_OPTION_WINDOW_SUPPORT);
            }
        }
        
        /* If XVideo is not being used, check if we're allowed to use MIT-SHM.
	 */

        if (!g_bUsingXVideoExtension) {
            /* Query Receiver as to whether SHM is being used. */

            g_bUsingXShmExtension = Get_SHM_event_base() > 0;
        }
    }


    if (g_bUsingXVideoExtension) {
        int          pix_fmt = 0x30323449; /* 'I420' (YUV420P) */
        unsigned int i;

        _XvQueryAdaptors(disp, DefaultRootWindow(disp),
			 &pCtxData->n_adaptors, &pCtxData->ai);
        
        if (pCtxData->ai) {
            /* Look for a suitable adaptor. */

            for (i = 0; i < pCtxData->n_adaptors; i++) {
                XvImageFormatValues *fmt_info;
	        int                  fmt_count;

                if (!(pCtxData->ai[i].type & XvImageMask)) {
                    continue;
                }
                fmt_info = _XvListImageFormats(disp, 
                                               pCtxData->ai[i].base_id,
                                               &fmt_count);
                if (fmt_info) {
                    for (int j = 0; j < fmt_count; j++) {                    
                        if (fmt_info[j].id == pix_fmt) {
                            goto found_adaptor;
                        }
                    }
                }
            }
found_adaptor:
            if (i < pCtxData->n_adaptors) {
                unsigned int j;

                /* Get a port. */

                for (j = 0; j < pCtxData->ai[i].num_ports; j++) {
                    pCtxData->xv_port = pCtxData->ai[i].base_id + j;
                    if (_XvGrabPort(disp, pCtxData->xv_port,
                                    CurrentTime) == Success) {
                        break;
                    }
                }
        
                if (j < pCtxData->ai[i].num_ports) {
                    pCtxData->fb_yuv =
                        _XvShmCreateImage(disp, pCtxData->xv_port, 
                                          pix_fmt, 0, width, height, 
                                          &pCtxData->shm_info);
                } else {
                    printf("No ports avaialble.\n");
                }
            } else {
                printf("No suitable adaptors.\n");
            }
            _XvFreeAdaptorInfo(pCtxData->ai);
        }

        if (pCtxData->fb_yuv) {
            size = pCtxData->fb_yuv->data_size;
        } else {
            /* Failed to create XvImage. Don't use XV. */

            g_bUsingXVideoExtension = false;
        }
    } else if (g_bUsingXShmExtension) {
        /* Use MIT-SHM. */

        pCtxData->fb = XShmCreateImage(disp, DefaultVisualOfScreen(scr),
                                       DefaultDepthOfScreen(scr), ZPixmap,
                                       NULL, &pCtxData->shm_info,
                                       width, height);
        if (pCtxData->fb) {
            size = pCtxData->fb->bytes_per_line * height;
        }
    } else {
        /* Use XPutImage. */

        pCtxData->fb = XCreateImage(disp, DefaultVisualOfScreen(scr),
                  DefaultDepthOfScreen(scr), ZPixmap, 0, NULL, width, height,
                  32, 0);
        if (pCtxData->fb) {
            size = pCtxData->fb->bytes_per_line * height;
        }
    }

    if (size > 0) {
        if (g_bUsingXVideoExtension || g_bUsingXShmExtension) {
            /* Allocated shared memory. */

            pCtxData->shm_info.shmseg = None;
            pCtxData->shm_info.readOnly = False;
            pCtxData->shm_info.shmid = shmget(IPC_PRIVATE, size,
                                              IPC_CREAT | 0600);
            if (pCtxData->shm_info.shmid == -1) {
	        delete pCtxData;
                return NULL;
            }
            pCtxData->shm_info.shmaddr =
                (char *)shmat(pCtxData->shm_info.shmid, 0, 0);
            if (pCtxData->shm_info.shmaddr == (void *)-1) {
	        delete pCtxData;
                return NULL;
            }

            /* Tell the X server to attach the segment. */

            if (XShmAttach(disp, &pCtxData->shm_info) > 0) {
                if (g_bUsingXVideoExtension) {
                    pCtxData->fb_yuv->data =
                        (char *)pCtxData->shm_info.shmaddr;
                } else {
                    pCtxData->fb->data = (char *)pCtxData->shm_info.shmaddr;
                }

                /* Mark the shared memory for removal now, so that it does
                 * not remain if this program dies unexpectedly.
                 */

                shmctl(pCtxData->shm_info.shmid, IPC_RMID, 0);
            }
        } else {
            /* Use traditional memory for XPutImage(). This memory is freed
             * by XDestroyImage();
             */

            pCtxData->fb->data = (char *)malloc(size);
        }
        // printf("xv = %d, xshm = %d\n",
        //        g_bUsingXVideoExtension, g_bUsingXShmExtension);
    } else {
      delete pCtxData;
      return NULL;
    }
#endif /* X11_SUPPORT */

    pCtxData->uEncFrameCurSize = 0;
    pCtxData->uEncFrameSize = 0;
    pCtxData->pEncodedFrame = NULL;

    g_GstCtxMap[CurCtx] = pCtxData;
    g_GstDecodeCtxListIdx ++;

#ifdef DUMP_OUTPUT
    {
        char str[256];

        sprintf(str, "/tmp/data_%d.h264", CurCtx);
        pCtxData->fp = fopen(str, "w+");
        if (pCtxData->fp) {
            fwrite(&width, 1, sizeof(int), pCtxData->fp);
            fwrite(&height, 1, sizeof(int), pCtxData->fp);
            fflush(pCtxData->fp);
        }
    }
#endif
    return pCtxData;
}

void v3_close_context(H264_context Ctx)
{
    V3FBCtxData *pCtxData = GetCtxData(Ctx);

    close_context(pCtxData);
    g_GstCtxMap.erase(Ctx);
}

/* When this returns non-NULL, mutex is still locked. */

V3FBCtxData *start_frame_common(H264_context Ctx, unsigned int encoded_size)
{
    V3FBCtxData *pCtxData;

    pCtxData = GetCtxData(Ctx);
    if (!pCtxData)
        return NULL;
    pthread_mutex_lock(&pCtxData->H264DataLock);
    pCtxData->uEncFrameCurSize = 0;
    pCtxData->uEncFrameSize = encoded_size;

    /* What sort of data is in this frame? Clear state variables. */

    pCtxData->small_frame = pCtxData->text_frame =
        pCtxData->h264_frame = false;

    if (encoded_size) {
        /* This frame contains H.264 data. */

        pCtxData->h264_frame = true;
        if (pCtxData->uEncFrameSize > pCtxData->uEncMaxFrameSize) {
            pCtxData->uEncMaxFrameSize = pCtxData->uEncFrameSize;
            if (pCtxData->pEncodedFrame) {
                free(pCtxData->pEncodedFrame);
            }

            pCtxData->pEncodedFrame =
	      (BYTE *)malloc(pCtxData->uEncMaxFrameSize);
	    if (!pCtxData->pEncodedFrame) {
                pthread_mutex_unlock(&pCtxData->H264DataLock);
                return NULL;
            }
        }
    } else {
        /* This is a small or text frame. */
    }
    return pCtxData;
}

void copy_dirty_region(V3FBCtxData *pCtxData,
		       SIGNED_RECT *dirty_rects, unsigned int num_rects)
{
    /* Save the dirty rects for this frame. */

    if (num_rects > 0) {
        for (unsigned int i = 0; i < num_rects; i++) {
            pCtxData->dirty_rects[i] = dirty_rects[i];
        }
        pCtxData->num_rects = num_rects;
    } else {
        /* No dirty rect means entire context needs updating. */

        pCtxData->dirty_rects[0].left = 0;
        pCtxData->dirty_rects[0].top = 0;
        pCtxData->dirty_rects[0].right = pCtxData->uCtxWidth;
        pCtxData->dirty_rects[0].bottom = pCtxData->uCtxHeight;
        pCtxData->num_rects = 1;
    }
}

/* When this returns non-NULL, mutex is still locked. */

V3FBCtxData *decode_frame_common(H264_context Ctx, void *H264_data,
                                 int len, bool last)
{
    V3FBCtxData *pCtxData;

    pCtxData = GetCtxData(Ctx);
    if (!pCtxData)
        return NULL;
    pthread_mutex_lock(&pCtxData->H264DataLock);
    memcpy(&pCtxData->pEncodedFrame[pCtxData->uEncFrameCurSize],
           H264_data, len);
    pCtxData->uEncFrameCurSize += len;

#ifdef DUMP_OUTPUT
    if (last) {
        if (pCtxData->fp) {
            fwrite(&pCtxData->uEncFrameCurSize, 1, sizeof (unsigned int),
                   pCtxData->fp);
            fwrite(pCtxData->pEncodedFrame, 1, pCtxData->uEncFrameCurSize,
                   pCtxData->fp);
            fflush(pCtxData->fp);
        }
    }
#endif
    return pCtxData;
}

bool v3_compose_with_fb(H264_context Ctx, ImgBuf* fb,
                        SIGNED_RECT interesting_rects[],
                        unsigned int num_rects)
{
    REFERENCE_VAR(interesting_rects);
    V3FBCtxData *pCtxData;
    unsigned int x, y;

    if (num_rects == 0) {
        return true;
    }

    pCtxData = GetCtxData(Ctx);
    pthread_mutex_lock(&pCtxData->H264DataLock);

    /* Very inefficient alpha blend. Assumes 32-bit LVB. */

    for (y = 0; y < fb->height; y++) {
        BYTE *alpha = (BYTE *)fb->bits + (y * fb->stride) + 3;

        for (x = 0; x < fb->width; x++) {
            if (H264_decoder.pref_lossless_alpha_val == *alpha) {
                /* Show this pixel. */

                BYTE *loc = (BYTE *)pCtxData->fb->data +
                    (y * pCtxData->fb->bytes_per_line) + (x * 4);
                *((UINT32 *)loc) = *((UINT32 *)(alpha - 3));
            }
            alpha += 4;
        }
    }

    pthread_mutex_unlock(&pCtxData->H264DataLock);
    return true;
}

static void free_lossless_object(struct image_buf &image)
{
    if (image.mem) {
#ifdef MAINLOOP
        struct special_buffer buf;

        if ((H264_decoder.options & H264_OPTION_PREFER_TEXT_RECTS) &&
            (Plug.special_allocate && 
             ((Plug.major_version > 1) ||
              (Plug.major_version == 1 && Plug.minor_version >= 2)))) {
            /* Allocated using "special_allocate()". So free
             * using "special_free()". Only supported for plugins with 
             * mainloop version >= 1.2.
             */

            buf.type = small_bitmap;
            buf.buf_addr = image.mem;
            Plug.special_free(&buf);
        } else
#endif
        {
            /* Allocated using traditional "malloc()". So
             * free using "free()".
             */

            free(image.mem);
        }
    }
}

bool v3_compose_with_rects(H264_context Ctx, ImgBuf rects[],
                           unsigned int num_rects, bool last)
{
    REFERENCE_VAR(last);
    V3FBCtxData* pCtxData;
    unsigned int i;

    if (num_rects == 0) {
        return true;
    }

    pCtxData = GetCtxData(Ctx);
    pthread_mutex_lock(&pCtxData->H264DataLock);

    for (i = 0; i < num_rects; i++) {
        if (IMAGE_OP_DRAW_LOSSLESS == rects[i].lossless_op ||
            IMAGE_OP_SMALL_FRAME_BITMAP == rects[i].lossless_op) {
            struct image_buf &image = rects[i];

            /* Take a copy of the image and store internally. */

            if (IMAGE_OP_DRAW_LOSSLESS == image.lossless_op) {
                /* Lossless text data. */        

                pCtxData->ltdata.push_back(image);
                
                /* This frame has *new* lossless text. */

                pCtxData->text_frame = true;
            } else {
                /* Small frame data, list to be purged on new H.264 frame. */

                pCtxData->sfdata.push_back(image);

                /* This frame has *new* small frame objects. */

                pCtxData->small_frame = true;
            }
        } else if (IMAGE_OP_DELETE_LOSSLESS == rects[i].lossless_op) {
            /* Find based on dest, width, height. This could be done better,
             * using a hash table.
             */

            for (std::vector<struct image_buf>::iterator it =
                     pCtxData->ltdata.begin();
                 it != pCtxData->ltdata.end();
                 it++) {
                struct image_buf &image = *it;

                if (   image.width == rects[i].width
                    && image.height == rects[i].height
                    && image.dst_x == rects[i].dst_x
                    && image.dst_y == rects[i].dst_y) {
                    /* Free lossless object. */

                    free_lossless_object(image);
                    pCtxData->ltdata.erase(it);
                    break;
                }
            }
        } else if (IMAGE_OP_SMALL_FRAME_SOLID_FILL == rects[i].lossless_op) {
            /* Small frame solid fill, list to be purged on new H.264 frame. */

            pCtxData->sfdata.push_back(rects[i]);

            /* This frame has *new* small frame objects. */

            pCtxData->small_frame = true;
        }
    }

    pthread_mutex_unlock(&pCtxData->H264DataLock);
    return true;
}

static bool intersects(SIGNED_RECT *r1, SIGNED_RECT *r2, SIGNED_RECT *r3)
{
    r3->top = max(r1->top, r2->top);
    r3->bottom = min(r1->bottom, r2->bottom);
    if (r3->top >= r3->bottom) {
        return false;
    }

    r3->left = max(r1->left, r2->left);
    r3->right = min(r1->right, r2->right);
    if (r3->left >= r3->right) {
        return false;
    }
    return true;
}

static void inflate_rect(SIGNED_RECT *rect, int amount)
{
    if (rect) {
        rect->left -= amount;
        rect->top -= amount;
        rect->right += amount;
        rect->bottom += amount;
    }
}

#ifdef CLIENT_INSTRUMENTATION
static void highlight_draw(V3FBCtxData *pCtxData, unsigned char lossless_op,
                           unsigned int x, unsigned int y,
                           unsigned int width, unsigned int height)
{
    /* Highlight if necessary. */

    if (cl_inst_flags & 0x800 /* CLIENT_INSTRUMENTATION_HIGHLIGHTING */) {
        UINT32       *save_loc, col;
        UINT32       *top_line, *left_line, *right_line, *bottom_line;
        int           stride;
        unsigned int  c;

        save_loc = (UINT32 *)((BYTE *)pCtxData->fb->data + 
                              (y * pCtxData->fb->bytes_per_line) + (x * 4));
        stride = pCtxData->fb->bytes_per_line >> 2;
        switch (lossless_op) {
            case IMAGE_OP_SMALL_FRAME_BITMAP:
                col = 0x000000FF; /* Blue. */
                break;            
            case IMAGE_OP_DRAW_LOSSLESS:
                col = 0x00FF0000; /* Red. */
                break;
            case IMAGE_OP_SMALL_FRAME_SOLID_FILL:
                col = 0x00FFFF00; /* Yellow. */
                break;
            case 0xFF: /* H.264 dirty rect push. */
                col = 0x0000FF00; /* Green. */
                break;
            default:
                col = 0x00000000; /* Black - shouldn't get here! */
                break;
        }

        top_line = save_loc;
        left_line = save_loc;
        right_line = save_loc + (width - 1);
        bottom_line = save_loc + (stride * (height - 1));
 
        /* Draw top and bottom lines. */

        for (c = 0; c < width; c++) {
            *top_line++ = col;
            *bottom_line++ = col;
        }
        /* Draw left and right lines. */

        for (c = 0; c < height; c++) {
            *left_line = col;
            left_line += stride;
            *right_line = col;
            right_line += stride;
        }     
    }
}
#endif

static void blt_object(V3FBCtxData *pCtxData, struct image_buf &image)
{
    BYTE *loc, *bits;

    loc = (BYTE *)pCtxData->fb->data +
                  (image.dst_y * pCtxData->fb->bytes_per_line) +
                  (image.dst_x * 4);
    bits = (BYTE *)image.bits;

    /* Make sure we draw in the pixel format we asked for. */

    if (image.pixel_format == H264_decoder.pref_lossless_pixel_fmt) {
        /* memcpy() each scanline onto the frame buffer. */

        for (unsigned int y = 0; y < image.height; y++) {
            memcpy(loc, bits, image.stride);
            loc += pCtxData->fb->bytes_per_line;
            bits += image.stride;
        }
    }

#ifdef CLIENT_INSTRUMENTATION
    highlight_draw(pCtxData, image.lossless_op,
                   image.dst_x, image.dst_y, image.width, image.height);
#endif
}

bool v3_push_frame(H264_context Ctx, struct window_info windows[],
                   unsigned int num_windows, bool wait, bool *pushed)
{
    V3FBCtxData *pCtxData;
#ifdef X11_SUPPORT
    static GC gc = None;

    if (gc == None) {
        /* Create a re-usable graphics context. */

        gc = XCreateGC(disp, DefaultRootWindow(disp), 0, 0);
    }
#endif /* X11_SUPPORT */

    pCtxData = GetCtxData(Ctx);

    /* Guard against NewH264FrameCB function operating on pH264Data. */

    pthread_mutex_lock(&pCtxData->H264DataLock);

    if ((H264_decoder.options & (H264_OPTION_PREFER_TEXT_RECTS |
                                 H264_OPTION_LOSSLESS))
        == (H264_OPTION_PREFER_TEXT_RECTS | H264_OPTION_LOSSLESS)) {
        /* Apply individual text rectangles if this is frame
         * contains H.264 data as the frame buffer will have been
         * overwritten with new H.264 data.
         *
         * Also, reapply text rectangles if we've received new text rectangles
         * for this frame.
         */

        if (pCtxData->h264_frame || pCtxData->text_frame) {
            /* Apply any text rectangles. */

            for (std::vector<struct image_buf>::iterator it =
                     pCtxData->ltdata.begin();
                 it != pCtxData->ltdata.end();
                 it++) {
                blt_object(pCtxData, *it);
            }
        }

        /* Apply new small frame objects if we have any. */

        if (pCtxData->small_frame) {
            for (std::vector<struct image_buf>::iterator it =
                     pCtxData->sfdata.begin();
                 it != pCtxData->sfdata.end();
                 it++) {
                struct image_buf &image = *it;

                if (IMAGE_OP_SMALL_FRAME_BITMAP == image.lossless_op) {
                    /* Apply any small frame objects to the frame buffer. */

                    blt_object(pCtxData, image);
                    /* As we've copied the data onto the frame buffer, we
                     * can now free it.
                     */

                    free_lossless_object(image);
                } else if (IMAGE_OP_SMALL_FRAME_SOLID_FILL
                             == image.lossless_op) {
                    BYTE *loc;

                    loc = (BYTE *)pCtxData->fb->data +
                              (image.dst_y * pCtxData->fb->bytes_per_line) +
                              (image.dst_x * 4);
                    for (unsigned int y = 0; y < image.height; y++) {
                        UINT32 *start = (UINT32 *)loc;

                        for (unsigned int x = 0; x < image.width; x++) {
                            *start++ = image.col;
                        }
                        loc += pCtxData->fb->bytes_per_line;
                    } 
#ifdef CLIENT_INSTRUMENTATION
                    highlight_draw(pCtxData, IMAGE_OP_SMALL_FRAME_SOLID_FILL,
                                   image.dst_x, image.dst_y,
                                   image.width, image.height);
#endif
                }
            }
            /* Clear all objects. */

            pCtxData->sfdata.clear();
        }
    }

    /* Show composed frame buffer. We must work out, based on the dirty rects.
     * what portions of the window(s) require updating.
     */

#ifdef X11_SUPPORT
    for (unsigned int i = 0; i < num_windows; i++) {
        Window X_window;
        SIGNED_RECT &window_rect = windows[i].rect;

        X_window = windows[i].id;

        /* Check if any of the dirty rects lie within this window. */

        for (int j = 0; j < pCtxData->num_rects; j++) {
            SIGNED_RECT dirty_rect, overlap;

            /* Inflate the dirty rect to minimize draw artifacts. */

            dirty_rect = pCtxData->dirty_rects[j];
            inflate_rect(&dirty_rect, 2);

            if (intersects(&window_rect, &dirty_rect, &overlap)) {
                int width, height, src_x, src_y, dest_x, dest_y;

                /* Don't use XvShmPutImage for seamless windows. */

                if (g_bUsingXVideoExtension || g_bUsingXShmExtension) {
                    if (g_bUsingXVideoExtension) {
                        /* Some alignment is required here for buggy video
                         * drivers. */

                        overlap.left &= ~1;
                        overlap.top &= ~1;
                        overlap.right = (overlap.right + 1) & ~1;
                        overlap.bottom = (overlap.bottom + 1) & ~1;

                        width = overlap.right - overlap.left;
                        height = overlap.bottom - overlap.top;

                        src_x = overlap.left;
                        src_y = overlap.top;

                        dest_x = (windows[i].target_x + overlap.left) & ~1;
                        dest_y = (windows[i].target_y + overlap.top) & ~1;

                        _XvShmPutImage(disp, pCtxData->xv_port, X_window,
                                       gc, pCtxData->fb_yuv,
                                       src_x, src_y, width, height,
                                       dest_x, dest_y, width, height,
                                       true);
                    } else {
                        /* No special alignment required. */

                        width = overlap.right - overlap.left;
                        height = overlap.bottom - overlap.top;

                        src_x = overlap.left;
                        src_y = overlap.top;

                        dest_x = windows[i].target_x + overlap.left;
                        dest_y = windows[i].target_y + overlap.top;

#ifdef CLIENT_INSTRUMENTATION
                        /* Highlight this draw if needed. */

                        if (pCtxData->h264_frame) {
                            int t_x, t_y;

                            /* Ensure highlighting occurs on the
                             * correct context buffer.
                             */

                            t_x = dest_x - windows[i].target_x;
                            t_y = dest_y - windows[i].target_y;
                            highlight_draw(pCtxData, 0xFF, t_x, t_y,
                                           width, height);
                        }
#endif
                        XShmPutImage(disp, X_window, gc, pCtxData->fb,
                                     src_x, src_y, dest_x, dest_y,
                                     width,  height, 
                                     true);
                    }

                    /* Flush the rendering pipeline and
                     * wait for synchronization.
                     */

                    /* ALWAYS exported by the main process. */

                    SHM_FramesPushed++;

                    if (wait) {
                        XFlush(disp);

                        /* Explicitly call WaitForShmCompletion() here. */

                        WaitForShmCompletion(); 
                    }
                } else {
                    /* XPutImage - no special alignment required. */

                    width = overlap.right - overlap.left;
                    height = overlap.bottom - overlap.top;
                    src_x = overlap.left;
                    src_y = overlap.top;

                    dest_x = windows[i].target_x + overlap.left;
                    dest_y = windows[i].target_y + overlap.top;

#ifdef CLIENT_INSTRUMENTATION
                    /* Highlight this draw if needed. */
                    if (pCtxData->h264_frame) {
                        int t_x, t_y;

                        /* Ensure highlighting occurs on
                         * the correct context buffer.
                         */

                        t_x = dest_x - windows[i].target_x;
                        t_y = dest_y - windows[i].target_y;
                        highlight_draw(pCtxData, 0xFF, t_x, t_y,
                                       width, height);
                    }
#endif
                    XPutImage(disp, X_window, gc, pCtxData->fb,
                              src_x,  src_y, dest_x, dest_y,
                              width, height);

                    /* No need to wait for syncrhonization with XPutImage(). */
                }
            }
        }
    }
#endif /* X11_SUPPORT */
    if (pushed) {
        *pushed = true;
    }

    pthread_mutex_unlock(&pCtxData->H264DataLock);
    return true;
}
